import argparse
import fcntl
import logging
import os
import sys
import tempfile
import threading
from pathlib import Path
from shutil import copy
from typing import List, Optional

import coloredlogs
import gi
import yaml

from .activity_reader import ActivityReader
from .application_info import ApplicationInfo
from .application_info_matcher import ApplicationInfoMatcher
from .application_info_reader import ApplicationInfoReader
from .config_reader import ConfigReader
from .dash_report_server import DashReportServer
from .files_provider import FilesProvider
from .localizator import Localizator
from .special_application_info_title import SpecialApplicationInfoTitle
from .special_wm_class import SpecialWmClass

gi.require_version('Wnck', '3.0')
gi.require_version('Gtk', '3.0')
gi.require_version('AppIndicator3', '0.1')
gi.require_version('Notify', '0.7')

from .speaking_eye_app import SpeakingEyeApp  # noqa: E402

APP_ID = 'speaking-eye'


def app_exit_with_failure(logger: logging.Logger, msg: str) -> None:
    logger.error(msg)
    sys.exit(1)


def dash_report_server_main(logger: logging.Logger,
                            config_reader: ConfigReader,
                            activity_reader: ActivityReader,
                            files_provider: FilesProvider,
                            localizator: Localizator) -> None:
    try:
        server = DashReportServer(logger, config_reader, activity_reader, files_provider, localizator)
        server.run()
    except Exception:
        logger.exception('Could not start Report Server!')


def make_autostartable_if_needed(config_reader: ConfigReader,
                                 files_provider: FilesProvider,
                                 logger: logging.Logger) -> None:
    autostart_file_path = files_provider.autostart_file_path

    if not config_reader.get_autostart():
        if autostart_file_path.exists():
            logger.debug(f'Remove [{autostart_file_path}] since [autostart=False] '
                         f'and file exists...')
            autostart_file_path.unlink()

        return

    if not autostart_file_path.exists():
        logger.debug(f'Copy [{files_provider.desktop_file_path}] into [{autostart_file_path}] '
                     f'since [autostart=True] and file does not exist...')
        copy(files_provider.desktop_file_path, autostart_file_path)


def main() -> None:
    current_file_dir = Path(os.path.dirname(os.path.abspath(__file__)))
    files_provider = FilesProvider(current_file_dir, APP_ID)
    default_config_path = str(files_provider.default_config_path)

    parser = argparse.ArgumentParser(description=f'[{APP_ID}] Track & analyze your computer activity')
    parser.add_argument('--log-level', type=str, choices=['debug', 'info', 'warning', 'error'],
                        default='debug', metavar='', help='debug/info/warning/error')
    parser.add_argument('-c', '--config', type=str, default=default_config_path, metavar='', help='config path')
    args = parser.parse_args()

    log_level_map = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR
    }

    # TODO: raise & exit if XDG_SESSION_TYPE is not x11 (check env vars for that)

    coloredlogs.install(log_level_map[args.log_level])
    logger = logging.getLogger(APP_ID)

    # TODO: It can be extracted into separate package with name like "run_once" -->
    pid_file = f'{tempfile.gettempdir()}/{APP_ID}.pid'
    fp = open(pid_file, 'w')

    try:
        fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        app_exit_with_failure(logger, msg='Another instance is already running!')
    # <--

    error_config_msg = 'Speaking Eye does not work without config. Bye baby!'

    try:
        files_provider.config_path = args.config
    except Exception:
        logger.exception(f'Error during config path [{args.config}] initialization!')
        app_exit_with_failure(logger, error_config_msg)

    logger.debug(f'Use config_path [{files_provider.config_path}]')

    config: Optional[ConfigReader.ConfigType] = None

    try:
        with open(str(files_provider.config_path)) as config_file:
            config = yaml.safe_load(config_file)
    except Exception:
        logger.exception(f'Config [{args.config}] is not correct')
        app_exit_with_failure(logger, error_config_msg)

    if not config:
        logger.error(f'Config [{args.config}] is empty')
        app_exit_with_failure(logger, error_config_msg)
        # NOTE: return is needed for correct mypy checks (config type)
        return

    application_info_reader = ApplicationInfoReader()
    config_reader = ConfigReader(application_info_reader, config)

    make_autostartable_if_needed(config_reader, files_provider, logger)

    detailed_app_infos: Optional[List[ApplicationInfo]] = None
    distracting_app_infos: Optional[List[ApplicationInfo]] = None

    try:
        detailed_app_infos = config_reader.try_read_application_info_list(ConfigReader.ConfigKey.DETAILED_NODE)
        distracting_app_infos = config_reader.try_read_application_info_list(
            ConfigReader.ConfigKey.DISTRACTING_NODE)
    except Exception:
        logger.exception('Error while reading config!')
        app_exit_with_failure(logger, error_config_msg)
        # NOTE: return is needed for correct mypy checks (detailed_app_infos & distracting_app_infos types)
        return

    # NOTE: Implicitly add LockScreen to monitor Break Time activity
    #       to not to add this info at user detailed apps config
    break_time_activity_info = ApplicationInfo(SpecialApplicationInfoTitle.BREAK_TIME.value,
                                               SpecialWmClass.LOCK_SCREEN.value,
                                               tab_re='',
                                               is_distracting=False)
    detailed_app_infos.append(break_time_activity_info)

    application_info_matcher = ApplicationInfoMatcher(detailed_app_infos, distracting_app_infos)
    activity_reader = ActivityReader(logger, application_info_matcher)

    language = config_reader.get_language()
    localizator = Localizator(files_provider.i18n_dir, language)
    logger.debug(f'Set user language to [{language.value}]')

    dash_server_thread = threading.Thread(target=dash_report_server_main,
                                          kwargs={
                                              'config_reader': config_reader,
                                              'logger': logger,
                                              'activity_reader': activity_reader,
                                              'files_provider': files_provider,
                                              'localizator': localizator,
                                          },
                                          daemon=True)
    dash_server_thread.start()

    app = SpeakingEyeApp(app_id=APP_ID,
                         config_reader=config_reader,
                         logger=logger,
                         application_info_matcher=application_info_matcher,
                         activity_reader=activity_reader,
                         files_provider=files_provider,
                         localizator=localizator)
    app.run()
    app.start_main_loop()


if __name__ == '__main__':
    main()



from .activity import Activity
from .bool_parser import BoolParser
from .datetime_formatter import DatetimeFormatter


class ActivityConverter:
    """Сonvert activity from/to string format"""

    EXPECTED_COLUMNS_NUMBER = 6

    @staticmethod
    def to_string(activity: Activity) -> str:
        return f'{activity.start_time}\t' \
               f'{activity.end_time}\t' \
               f'{activity.activity_time}\t' \
               f'{activity.wm_class}\t' \
               f'{activity.window_name}\t' \
               f'{activity.is_work_time}\n'

    @staticmethod
    def from_string(value: str) -> Activity:
        if value[-1] != '\n':
            raise ValueError(f'String [{value}] should be ended with a new line!')

        columns = value.split(sep='\t')

        if len(columns) != ActivityConverter.EXPECTED_COLUMNS_NUMBER:
            raise ValueError(f'Unexpected columns number: [{len(columns)}] '
                             f'!= {ActivityConverter.EXPECTED_COLUMNS_NUMBER} in string [{value}]!')

        try:
            start_time_str, end_time_str, activity_time, wm_class, window_name, is_work_time_str = columns
            start_time = DatetimeFormatter.parse_string_with_optional_milliseconds(start_time_str)
            end_time = DatetimeFormatter.parse_string_with_optional_milliseconds(end_time_str)
            is_work_time = BoolParser.parse(is_work_time_str.replace('\n', ''))
        except Exception as e:
            raise ValueError(f'Incorrect string [{value}]! {e}')

        return Activity(wm_class, window_name, start_time, is_work_time).set_end_time(end_time)



from datetime import datetime, timedelta


class DatetimeFormatter:
    """Format date/time objects as a strings or vice versa"""
    __TIME_SEPARATOR = ':'

    @staticmethod
    def __timedelta_to_datetime(value: timedelta) -> datetime:
        return datetime.min + value

    @staticmethod
    def __days_to_hours(value: int) -> int:
        return 24 * value

    @staticmethod
    def format_time_without_seconds(value: timedelta) -> str:
        """Take timedelta and return it as a string with hours and minutes"""
        time = DatetimeFormatter.__timedelta_to_datetime(value).time()
        hours = DatetimeFormatter.__days_to_hours(value.days) + time.hour

        return f'{hours}{DatetimeFormatter.__TIME_SEPARATOR}{time.minute:02}'

    @staticmethod
    def parse_string_with_optional_milliseconds(value: str) -> datetime:
        try:
            return datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            # NOTE: Datetime object such as datetime.now() that is saved as string
            #       when a new second just started, does not contain zero milliseconds after point
            return datetime.strptime(value, '%Y-%m-%d %H:%M:%S')



from datetime import date, timedelta, datetime
from typing import List, Tuple


DateRange = Tuple[
    datetime,  # start
    datetime,  # end
]


class DatetimeHelper:
    """Extra methods for working with date/time objects"""

    @staticmethod
    def get_dates_between(start: date, end: date) -> List[date]:
        if end < start:
            raise ValueError(f'End date [{end}] should be greater than start date [{start}]!')

        if start == end:
            return [start]

        delta = end - start

        return [start + timedelta(days=i) for i in range(delta.days + 1)]

    @staticmethod
    def get_date_range(date_: date) -> DateRange:
        """Get start & end point for the specific date"""
        start = datetime(date_.year, date_.month, date_.day)
        end = start + timedelta(days=1) - timedelta(microseconds=1)

        return start, end



from enum import Enum

from typing import Type, TypeVar

# declare type variable
DerivedEnumClass = TypeVar('DerivedEnumClass', bound=Enum)


# TODO: This class can be published into github as a pip package
#       because class is cool and it is a good idea for learning
class ExtendedEnum(Enum):
    """Used as a base class for enums with strict typing"""

    @classmethod
    def parse(cls: Type[DerivedEnumClass], value: str, default: DerivedEnumClass) -> DerivedEnumClass:
        """Parse string to enum member. Return default otherwise"""
        for enum_item in cls:
            if value == enum_item.value:
                return enum_item

        return default



import logging
import re
import signal
import subprocess
import webbrowser
from datetime import date, datetime, timedelta
from enum import Enum
from pathlib import Path
from random import choice
from types import FrameType
from typing import Any, cast, List, Optional

from gi.repository import Gio, GLib, GObject, Gtk, Notify, Wnck
from pyee import BaseEventEmitter

from .activity import Activity
from .activity_reader import ActivityReader
from .activity_stat_holder import ActivityStatHolder
from .activity_writer import ActivityWriter
from .application_info_matcher import ApplicationInfoMatcher
from .config_reader import ConfigReader
from .files_provider import FilesProvider
from .gtk_extras import get_window_name
from .icon_state import IconState
from .localizator import Localizator
from .notification import Notification, NotificationEvent
from .notification_emojis import BREAK_TIME_EMOJIS, DISTRACTING_NOTIFICATION_EMOJIS
from .special_wm_class import SpecialWmClass
from .timer import Timer
from .tray_icon import TrayIcon
from .value import Value
from .x_helpers import get_wm_class


class ApplicationEvent(Enum):
    DISTRACTING_APP_OVERTIME = 'distracting_app_overtime'


class SpeakingEyeApp(Gtk.Application):  # type: ignore[misc]

    def __init__(self,
                 app_id: str,
                 config_reader: ConfigReader,
                 logger: logging.Logger,
                 application_info_matcher: ApplicationInfoMatcher,
                 activity_reader: ActivityReader,
                 files_provider: FilesProvider,
                 localizator: Localizator) -> None:
        super().__init__()
        self.logger = logger
        self.config_reader = config_reader

        self.connection = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        self.screen_saver_bus_names = self.__dbus_get_screen_saver_bus_names()

        self.files_provider = files_provider
        self.localizator = localizator

        self.theme = self.config_reader.get_theme()
        self.active_icon = self.get_icon(IconState.ACTIVE)
        self.disabled_icon = self.get_icon(IconState.DISABLED)

        self.work_state_checkbox_item = self.__create_work_state_checkbox_item()
        self.tray_icon = TrayIcon(app_id, self.disabled_icon, self.create_tray_menu(self.work_state_checkbox_item))

        self.screen: Optional[Wnck.Screen] = None
        self.main_loop: Optional[GObject.MainLoop] = None
        self.name_changed_handler_id = None

        self.previous_active_window_name: Optional[str] = None
        self.previous_wm_class: Optional[str] = None

        self.overtime_timer = \
            Timer('overtime_timer', handler=self.overtime_timer_handler,
                  interval_ms=1 * 60 * 1000, repeat=True, logger=self.logger)
        self.break_timer = \
            Timer('break_timer', handler=self.break_timer_handler,
                  interval_ms=1 * 60 * 1000, repeat=True, logger=self.logger)
        self.distracting_app_timer = \
            Timer('distracting_app_timer', handler=self.distracting_app_timer_handler,
                  interval_ms=1 * 60 * 1000, repeat=True, logger=self.logger)

        self.event = BaseEventEmitter()

        self.event.on(ApplicationEvent.DISTRACTING_APP_OVERTIME.value, self.show_distracting_app_overtime_notification)

        self.start_time = datetime.now()
        self.is_work_time = False
        self.is_work_time_update_time = self.start_time
        self.last_overtime_notification: Optional[Notification] = None
        self.last_break_notification: Optional[Notification] = None
        self.last_lock_screen_time: Optional[datetime] = None
        self.is_lock_screen_activated = False
        self.has_distracting_app_overtime_notification_shown = False
        self.is_overtime_notification_allowed_to_show = True
        self.is_break_notification_allowed_to_show = True

        self.user_work_time_hour_limit = config_reader.get_work_time_limit()
        self.user_breaks_interval_hours = config_reader.get_breaks_interval_hours()
        self.user_distracting_apps_mins = config_reader.get_distracting_apps_mins()

        self.writer = ActivityWriter(self.files_provider)

        self.app_info_matcher = application_info_matcher

        today_raw_data_file_path = self.files_provider.get_raw_data_file_path(date.today())
        self.holder = ActivityStatHolder(activity_reader.read(today_raw_data_file_path))

        self.holder.initialize_stats(self.app_info_matcher.detailed_app_infos)
        self.holder.initialize_stats(self.app_info_matcher.distracting_app_infos)

        self.current_activity: Optional[Activity] = None

        self.writer.event.on(ActivityWriter.NEW_DAY_EVENT, self.__on_new_day_started)

        self.logger.debug(f'Set user work time limit to [{self.user_work_time_hour_limit}] hours')
        self.logger.debug(f'Set user user breaks interval to [{self.user_breaks_interval_hours}] hours')

        Notify.init(app_id)
        self.__dbus_subscribe_to_screen_saver_signals()

        start_msg = self.localizator.get('notification.start', start_time=self.start_time.strftime("%H:%M:%S"))
        self.logger.debug(start_msg)
        self.new_notification(msg=start_msg).show()

    def __dbus_method_call(self, bus_name: str, object_path: str, interface_name: str, method_name: str) -> Any:
        """
        D-Bus is a middleware mechanism that allows communication between
        multiple processes running concurrently on the same machine
        """
        if not self.connection:
            raise Exception('self.connection should be set!')

        no_parameters = None
        default_reply_type = None
        default_call_timeout = -1
        not_cancellable = None

        raw_result = self.connection.call_sync(bus_name, object_path, interface_name,
                                               method_name, no_parameters, default_reply_type,
                                               Gio.DBusCallFlags.NONE, default_call_timeout, not_cancellable)

        if raw_result:
            result, = raw_result

            return result

        return None

    def __dbus_get_all_bus_names(self) -> List[str]:
        return cast(
            List[str],
            self.__dbus_method_call('org.freedesktop.DBus', '/org/freedesktop/DBus',
                                    'org.freedesktop.DBus', 'ListNames')
        )

    def __dbus_lock_screen(self) -> None:
        """Try to lock screen with a dbus call"""
        for bus in self.screen_saver_bus_names:
            if bus == 'org.freedesktop.ScreenSaver':
                # NOTE: that server is just interface without implementation
                continue

            interface_name = f'/{bus.replace(".", "/")}'

            try:
                self.__dbus_method_call(bus, interface_name, bus, 'Lock')
            except Exception as e:
                self.logger.warning(f'Please ignore it if lock screen works well. Lock screen error: [{e}]')

    def __on_screen_saver_active_changed(self, connection: Gio.DBusConnection, sender_name: str, object_path: str,
                                         interface_name: str, signal_name: str, parameters: GLib.Variant) -> None:
        is_activated, = parameters
        self.is_lock_screen_activated = is_activated

        now = datetime.now()

        if self.is_lock_screen_activated:
            current_activity = Value.get_or_raise(self.current_activity, 'current_activity')

            self.previous_wm_class = current_activity.wm_class
            self.previous_active_window_name = current_activity.window_name

            wm_class = SpecialWmClass.LOCK_SCREEN.value
            window_name = ''
        else:
            # the screen has just been unlocked
            if self.is_work_time:
                self.last_lock_screen_time = now

            wm_class = Value.get_or_raise(self.previous_wm_class, 'previous_wm_class')
            window_name = Value.get_or_raise(self.previous_active_window_name, 'previous_active_window_name')

        self.__on_open_window(wm_class, window_name, now)

    def __dbus_get_screen_saver_bus_names(self) -> List[str]:
        bus_names = self.__dbus_get_all_bus_names()
        screen_saver_re = re.compile(r'^org\..*\.ScreenSaver$')

        return list(filter(screen_saver_re.match, bus_names))

    def __dbus_subscribe_to_screen_saver_signals(self) -> None:
        if not self.screen_saver_bus_names:
            raise Exception('self.screen_saver_bus_names should be set!')

        for bus_name in self.screen_saver_bus_names:
            self.connection.signal_subscribe(None, bus_name, 'ActiveChanged', None, None,
                                             Gio.DBusSignalFlags.NONE, self.__on_screen_saver_active_changed)

    def do_activate(self) -> None:
        """Gtk.Application must call this method to get the application up and running"""
        signal.signal(signal.SIGTERM, self.handle_sigterm)
        self.screen = Wnck.Screen.get_default()
        self.screen.connect('active-window-changed', self.__on_active_window_changed)
        self.main_loop = GObject.MainLoop()
        self.overtime_timer.start()
        self.break_timer.start()
        self.distracting_app_timer.start()

    def __on_active_window_changed(self, screen: Wnck.Screen, previously_active_window: Gtk.Window) -> None:
        now = datetime.now()

        # to prevent double handler connections
        if previously_active_window and self.name_changed_handler_id:
            previously_active_window.disconnect(self.name_changed_handler_id)

        active_window = screen.get_active_window()

        if active_window:
            self.name_changed_handler_id = active_window.connect('name-changed', self.__on_name_changed)
            wm_class = get_wm_class(active_window.get_xid())
            window_name = get_window_name(active_window)
        else:
            wm_class = SpecialWmClass.DESKTOP.value
            window_name = ''

        self.__on_open_window(wm_class, window_name, now)

    def __on_open_window(self, wm_class: str, window_name: str, now: datetime) -> None:
        new_activity = Activity(wm_class, window_name, now, self.is_work_time)

        self.__on_activity_changed(self.current_activity, new_activity)

    def __on_name_changed(self, window: Wnck.Window) -> None:
        now = datetime.now()

        current_activity = Value.get_or_raise(self.current_activity, 'current_activity')
        window_name = get_window_name(window)

        new_activity = Activity(current_activity.wm_class, window_name, now, current_activity.is_work_time)

        self.__on_activity_changed(current_activity, new_activity)

    def __on_activity_changed(self, previous_activity: Optional[Activity], next_activity: Activity) -> None:
        now = datetime.now()

        if previous_activity is not None:
            previous_activity.set_end_time(now)
            self.writer.write(previous_activity)
            self.holder.update_stat(previous_activity)

        # NOTE: previous_activity is None when it is the first activity after starting
        previous_activity_app_name = \
            '' if previous_activity is None else f'{previous_activity.wm_class}|{previous_activity.window_name}'
        self.logger.debug(f'{now}: {previous_activity_app_name} -> '
                          f'{next_activity.wm_class}|{next_activity.window_name}')

        self.app_info_matcher.set_if_matched(next_activity)

        self.current_activity = next_activity
        # NOTE: for reshowing notification when open distracting app once more time
        self.has_distracting_app_overtime_notification_shown = False

    def start_main_loop(self) -> None:
        try:
            self.main_loop.run()  # type: ignore[union-attr]
        except KeyboardInterrupt:
            self.stop()

    def stop(self) -> None:
        """Stop application work and quit main loop"""
        now = datetime.now()

        finish_time = now
        work_time = finish_time - self.start_time
        work_time -= timedelta(microseconds=work_time.microseconds)

        finish_msg = self.localizator.get('notification.finish', finish_time=finish_time.strftime("%H:%M:%S"))
        work_time_msg = self.localizator.get('notification.work_time', work_time=work_time)

        self.logger.debug(f'{finish_msg}\n{work_time_msg}')

        self.logger.info('              title |          work_time |            off_time')
        self.logger.info('--------------------------------------------------------------')

        for holder_item in self.holder.items():
            title, stat = holder_item

            padded_title = title.rjust(19, ' ')
            padded_work_time = f'{stat.work_time}'.rjust(19, ' ')
            padded_off_time = f'{stat.off_time}'.rjust(20, ' ')
            self.logger.info(f'{padded_title} |{padded_work_time} |{padded_off_time}')

        self.new_notification(msg=f'{finish_msg}; {work_time_msg}').show()
        Notify.uninit()

        self.main_loop.quit()  # type: ignore[union-attr]

    def __on_close_item_click(self, menu_item: Gtk.MenuItem) -> None:
        self.stop()

    def __on_open_report_item_click(self, menu_item: Gtk.MenuItem) -> None:
        """Open page with report in browser"""
        browser = webbrowser.get(self.config_reader.get_report_server_browser())
        host = self.config_reader.get_report_server_host()
        port = self.config_reader.get_report_server_port()
        url = f'http://{host}:{port}'

        browser.open_new_tab(url)

    def __on_edit_config_item_click(self, menu_item: Gtk.MenuItem) -> None:
        subprocess.run(['xdg-open', f'{self.files_provider.config_path}'])

    def __on_open_data_dir_item_click(self, menu_item: Gtk.MenuItem) -> None:
        subprocess.run(['xdg-open', f'{self.files_provider.raw_data_dir}'])

    def set_work_time_state(self, value: bool) -> None:
        """
        Try to change work time state. It can be True (working time) or False (not working time).
        If changed successfully, tray icon & work state checkbox in menu should also change
        """
        if value == self.is_work_time:
            self.logger.debug('Trying to change is_work_time to the same value')
            return

        self.is_work_time = value
        self.work_state_checkbox_item.set_active(self.is_work_time)

        current_activity = Value.get_or_raise(self.current_activity, 'current_activity')
        now = datetime.now()
        new_activity = Activity(current_activity.wm_class,
                                current_activity.window_name,
                                now,
                                self.is_work_time)

        self.__on_activity_changed(current_activity, new_activity)

        self.is_work_time_update_time = now

        self.logger.debug(f'Set Work Time to [{self.is_work_time}]')

        icon = self.active_icon if self.is_work_time else self.disabled_icon
        self.tray_icon.set_icon_if_exist(icon)

        if self.is_work_time:
            self.last_lock_screen_time = now
            self.last_break_notification = None
            self.last_overtime_notification = None

    def on_work_state_checkbox_item_click(self, menu_item: Gtk.CheckMenuItem) -> None:
        """Reverse work time state"""
        self.set_work_time_state(not self.is_work_time)

    def __create_work_state_checkbox_item(self) -> Gtk.CheckMenuItem:
        work_state_checkbox_item = Gtk.CheckMenuItem(self.localizator.get('tray.work_time'))
        work_state_checkbox_item.connect('activate', self.on_work_state_checkbox_item_click)

        return work_state_checkbox_item

    def create_tray_menu(self, work_state_checkbox_item: Gtk.CheckMenuItem) -> Gtk.Menu:
        menu = Gtk.Menu()

        menu.append(work_state_checkbox_item)

        open_report_item = Gtk.MenuItem(self.localizator.get('tray.open_report'))
        open_report_item.connect('activate', self.__on_open_report_item_click)
        menu.append(open_report_item)

        menu.append(Gtk.SeparatorMenuItem())

        edit_config_item = Gtk.MenuItem(self.localizator.get('tray.edit_config'))
        edit_config_item.connect('activate', self.__on_edit_config_item_click)
        menu.append(edit_config_item)

        open_data_dir_item = Gtk.MenuItem(self.localizator.get('tray.open_data_dir'))
        open_data_dir_item.connect('activate', self.__on_open_data_dir_item_click)
        menu.append(open_data_dir_item)

        menu.append(Gtk.SeparatorMenuItem())

        close_item = Gtk.MenuItem(self.localizator.get('tray.close'))
        close_item.connect('activate', self.__on_close_item_click)
        menu.append(close_item)

        menu.show_all()

        return menu

    def __on_overtime_notification_closed(self) -> None:
        self.is_overtime_notification_allowed_to_show = True

    def __on_finish_work_action_clicked(self) -> None:
        self.set_work_time_state(False)

    def __on_take_break_clicked(self) -> None:
        self.__dbus_lock_screen()

    def new_notification(self,
                         msg: str,
                         urgency: Optional[Notify.Urgency] = None,
                         listen_closed_event: bool = False) -> Notification:
        # TODO: replace with NotificationFactory (Factory Pattern)
        return Notification('Speaking Eye', msg, self.active_icon, urgency, listen_closed_event)

    def show_overtime_notification(self) -> None:
        msg = self.localizator.get('notification.overtime.text', hours=self.user_work_time_hour_limit)

        notification = self.new_notification(msg, Notify.Urgency.CRITICAL, listen_closed_event=True)

        notification.add_buttons((
            self.localizator.get('notification.overtime.left_button'),  # finish_work
            self.localizator.get('notification.overtime.right_button')  # remind_later
        ))

        notification.event.on(NotificationEvent.CLOSED.value, self.__on_overtime_notification_closed)
        notification.event.on(NotificationEvent.LEFT_BUTTON_CLICKED.value, self.__on_finish_work_action_clicked)

        notification.show()

        self.last_overtime_notification = notification
        self.is_overtime_notification_allowed_to_show = False

    def show_distracting_app_overtime_notification(self, title: str, total_time: timedelta) -> None:
        distracting_minutes = total_time.total_seconds() // 60
        text = self.localizator.get_random('notification.distracting_texts', 10)
        emoji = choice(DISTRACTING_NOTIFICATION_EMOJIS)
        msg = self.localizator.get('notification.distracting', app_title=title,
                                   distracting_minutes=int(distracting_minutes),
                                   text=text, emoji=emoji)

        self.new_notification(msg).show()

    def __on_break_notification_closed(self) -> None:
        self.is_break_notification_allowed_to_show = True

    def show_break_notification(self) -> None:
        emoji = choice(BREAK_TIME_EMOJIS)
        msg = self.localizator.get('notification.break.text', emoji=emoji)

        notification = self.new_notification(msg, Notify.Urgency.CRITICAL, listen_closed_event=True)

        notification.add_buttons((
            self.localizator.get('notification.break.left_button'),  # take_break
            self.localizator.get('notification.break.right_button')  # remind_later
        ))

        notification.event.on(NotificationEvent.CLOSED.value, self.__on_break_notification_closed)
        notification.event.on(NotificationEvent.LEFT_BUTTON_CLICKED.value, self.__on_take_break_clicked)

        notification.show()

        self.last_break_notification = notification
        self.is_break_notification_allowed_to_show = False

    def __on_new_day_started(self) -> None:
        """Reset work time state"""
        open_new_file_msg = self.localizator.get('notification.new_day')

        self.logger.debug(open_new_file_msg)
        self.new_notification(msg=open_new_file_msg).show()

        self.set_work_time_state(False)

    def handle_sigterm(self, signal_number: int, frame: FrameType) -> None:
        self.stop()

    def __need_to_show_overtime_notification(self) -> bool:
        if not self.is_work_time:
            return False

        if self.is_lock_screen_activated:
            return False

        now = datetime.now()
        current_activity = Value.get_or_raise(self.current_activity, 'current_activity')
        seconds_in_current_activity = (now - current_activity.start_time).total_seconds()
        total_work_time_seconds = seconds_in_current_activity + self.holder.total_work_time.total_seconds()
        is_overtime_started = total_work_time_seconds >= self.user_work_time_hour_limit * 60 * 60

        if not is_overtime_started:
            return False

        if self.last_overtime_notification is None:
            return True

        if self.last_overtime_notification.last_shown is None:
            return True

        if not self.is_overtime_notification_allowed_to_show:
            return False

        seconds_from_last_notification = (now - self.last_overtime_notification.last_shown).total_seconds()
        interval_seconds = 15 * 60

        return seconds_from_last_notification >= interval_seconds

    def overtime_timer_handler(self) -> None:
        if not self.__need_to_show_overtime_notification():
            return

        self.show_overtime_notification()

    def __need_to_show_break_notification(self) -> bool:
        if not self.is_work_time:
            return False

        if self.is_lock_screen_activated:
            return False

        if not self.is_break_notification_allowed_to_show:
            return False

        now = datetime.now()
        start_work_time = self.is_work_time_update_time
        last_break_time = self.last_lock_screen_time if self.last_lock_screen_time else start_work_time

        if (now - last_break_time).total_seconds() < self.user_breaks_interval_hours * 60 * 60:
            return False

        last_break_reminder_time = Value.get_by_getter_or_default(
            lambda: self.last_break_notification.last_shown,  # type: ignore[union-attr]
            start_work_time
        )

        if (now - last_break_reminder_time).total_seconds() < 15 * 60:
            return False

        return True

    def break_timer_handler(self) -> None:
        if not self.__need_to_show_break_notification():
            return

        self.show_break_notification()

    def distracting_app_timer_handler(self) -> None:
        # TODO: ⚡️ run this timer only if work started
        if not self.is_work_time:
            return

        if self.has_distracting_app_overtime_notification_shown:
            return

        current_activity = Value.get_or_raise(self.current_activity, 'current_activity')
        application_info = current_activity.application_info

        if application_info is None:
            # NOTE: It is None if detailed/distracting lists do not contain such activity
            #       See more in ApplicationInfoMatcher.set_if_matched()
            return

        if not application_info.is_distracting:
            return

        now = datetime.now()
        current_stats = self.holder[application_info.title]
        total_distracting_time = now - current_activity.start_time + current_stats.work_time

        if total_distracting_time.total_seconds() < self.user_distracting_apps_mins * 60:
            return

        self.event.emit(ApplicationEvent.DISTRACTING_APP_OVERTIME.value, application_info.title, total_distracting_time)

        self.has_distracting_app_overtime_notification_shown = True

    def get_icon(self, icon_state: IconState) -> Path:
        if not self.theme:
            raise Exception('self.theme should be set!')

        path = self.files_provider.get_icon_file_path(self.theme, icon_state)

        if not path.exists():
            self.logger.warning(f'Icon [{path}] not found')

        return path



from enum import Enum
from typing import Any, List, Dict, Optional

from pydash import get

from .application_info import ApplicationInfo
from .application_info_reader import ApplicationInfoReader
from .language import Language
from .theme import Theme
from .typed_value import TypedValue


class ConfigReader:
    """
    Parse config and get values from it.
    Allows to create application info lists based on config nodes
    """
    ConfigType = Dict[str, Any]

    class ConfigKey(Enum):
        DETAILED_NODE = 'detailed'
        DISTRACTING_NODE = 'distracting'

    def __init__(self,
                 application_info_reader: ApplicationInfoReader,
                 config: ConfigType) -> None:
        self.application_info_reader = application_info_reader
        self.config = config

    def try_read_application_info_list(self, config_key: ConfigKey) -> List[ApplicationInfo]:
        """Try to read detailed/distracting ApplicationInfos from app config"""
        apps_list_node = f'apps.{config_key.value}'

        # TODO: get with TypedValue.get() to check types
        app_list = get(self.config, apps_list_node)

        is_distracting = config_key == self.ConfigKey.DISTRACTING_NODE

        if app_list is not None:
            return self.application_info_reader.try_read(app_list, is_distracting)

        if is_distracting:
            return []

        raise RuntimeError(f'Path [{apps_list_node}] should be set in config!')

    def get_work_time_limit(self) -> int:
        return TypedValue.get(self.config, 'time_limits.work_time_hours', int, 9)

    def get_breaks_interval_hours(self) -> int:
        return TypedValue.get(self.config, 'time_limits.breaks_interval_hours', int, 3)

    def get_distracting_apps_mins(self) -> int:
        return TypedValue.get(self.config, 'time_limits.distracting_apps_mins', int, 15)

    def get_report_server_host(self) -> str:
        return TypedValue.get(self.config, 'report_server.host', str, 'localhost')

    def get_report_server_port(self) -> int:
        return TypedValue.get(self.config, 'report_server.port', int, 3838)

    def get_report_server_browser(self) -> Optional[str]:
        return TypedValue.get(self.config, 'report_server.browser', Optional[str], None)  # type: ignore[arg-type]

    def get_report_server_ignore_weekends(self) -> bool:
        return TypedValue.get(self.config, 'report_server.ignore_weekends', bool, True)

    def get_language(self) -> Language:
        return Language.parse(get(self.config, 'language'), Language.ENGLISH)

    def get_theme(self) -> Theme:
        return Theme.parse(get(self.config, 'theme'), Theme.DARK)

    def get_autostart(self) -> bool:
        return TypedValue.get(self.config, 'autostart', bool, True)



from re import search
from typing import List, Optional

from .activity import Activity
from .application_info import ApplicationInfo


class ApplicationInfoMatcher:
    """
    Match ApplicationInfo with Activity
    to distinguish between different types of activities/infos (e.g. 'detailed/distracting')
    """

    def __init__(self, detailed_app_infos: List[ApplicationInfo], distracting_app_infos: List[ApplicationInfo]) -> None:
        self.detailed_app_infos = detailed_app_infos
        self.distracting_app_infos = distracting_app_infos

    def __try_find_app_info_in(self, app_infos: List[ApplicationInfo], activity: Activity) -> Optional[ApplicationInfo]:
        """Using Activity wm_class and window_name try to find corresponding ApplicationInfo"""

        for app_info in app_infos:
            # tab can be empty str, it means that we are considering all tabs
            # search('', whatever) always return SRE_Match object
            if search(app_info.wm_name_re, activity.wm_class) and search(app_info.tab_re, activity.window_name):
                return app_info

        return None

    def set_if_matched(self, activity: Activity) -> None:
        """Set ApplicationInfo to Activity if matched"""

        application_info = self.__try_find_app_info_in(
            self.distracting_app_infos, activity) or self.__try_find_app_info_in(self.detailed_app_infos, activity)

        if application_info is None:
            return

        activity.set_application_info(application_info)



from typing import Optional, cast

from gi.repository import Gtk


def get_window_name(window: Optional[Gtk.Window]) -> str:
    """Get window name using GTK or label for empty otherwise"""
    return cast(str, window.get_name()) if window else 'EMPTY WINDOW'



from datetime import datetime
from typing import Optional

from .activity import Activity


class ActivityBuilder:
    """Implementation of the Builder pattern for easy building Activity objects"""

    def __init__(self, origin_activity: Activity) -> None:
        self.origin_activity = origin_activity
        self.start_time: Optional[datetime] = None
        self.end_time: Optional[datetime] = None

    def with_start_time(self, value: datetime) -> 'ActivityBuilder':
        self.start_time = value

        return self

    def with_end_time(self, value: datetime) -> 'ActivityBuilder':
        self.end_time = value

        return self

    def build(self) -> Activity:
        if self.start_time is None:
            raise RuntimeError('start_time should be set!')

        if self.end_time is None:
            raise RuntimeError('end_time should be set!')

        activity = Activity(
            self.origin_activity.wm_class,
            self.origin_activity.window_name,
            self.start_time,
            self.origin_activity.is_work_time
        )

        activity.set_end_time(self.end_time)

        return activity



import logging
import re
from datetime import date, datetime, timedelta
from enum import Enum
from random import choice
from typing import List, Optional, Tuple

import dash_core_components as dcc
import dash_html_components as html
import pandas as pd
import plotly.express as px
from dash import Dash
from dash.dependencies import Input, Output
from pydash import some

from .activity_reader import ActivityReader
from .activity_stat_holder import ActivityStatHolder
from .config_reader import ConfigReader
from .datetime_formatter import DatetimeFormatter
from .datetime_helper import DatetimeHelper
from .files_provider import FilesProvider
from .localizator import Localizator
from .special_application_info_title import SpecialApplicationInfoTitle

COLORS_SEQUENTIALS = [
    px.colors.sequential.Viridis,
    px.colors.sequential.GnBu,
    px.colors.sequential.deep,
    px.colors.sequential.dense,
]


class ElementId(Enum):
    DATE_PICKER = 'date-picker'
    REPORT_OUTPUT = 'report-output'


GetActivityStatHolderResultType = Tuple[
    ActivityStatHolder,
    int,  # active_days_count
]


class DashReportServer:

    class ColumnName(Enum):
        TITLE = 'title'
        WORK_TIME = 'work_time'
        MEAN_WORK_TIME = 'mean_work_time'
        MEAN_WORK_TIME_STR = 'mean_work_time_str'

    def __init__(self,
                 logger: logging.Logger,
                 app_config_reader: ConfigReader,
                 activity_reader: ActivityReader,
                 files_provider: FilesProvider,
                 localizator: Localizator) -> None:
        self.app = Dash(__name__, assets_folder='./assets')

        logging.getLogger('werkzeug').setLevel(logging.WARNING)
        logging.getLogger('parse').setLevel(logging.WARNING)
        self.app.logger = logger

        self.localizator = localizator
        self.app.title = self.localizator.get('report_server.tab_title')

        self.host = app_config_reader.get_report_server_host()
        self.port = app_config_reader.get_report_server_port()

        self.work_time_limit = app_config_reader.get_work_time_limit()
        self.distracting_apps_mins = app_config_reader.get_distracting_apps_mins()

        self.ignore_weekends = app_config_reader.get_report_server_ignore_weekends()

        self.activity_reader = activity_reader
        self.files_provider = files_provider

        self.colors = choice(COLORS_SEQUENTIALS)

        self.app.layout = self.__get_layout()

    def __get_activity_stat_holder(self, report_dates: List[date]) -> GetActivityStatHolderResultType:
        """Get ActivityStatHolder with all collected activities for specific dates"""
        active_days_count = 0
        all_activities = []
        for report_date in report_dates:
            is_weekend = report_date.weekday() > 4

            if self.ignore_weekends and is_weekend:
                continue

            file_path = self.files_provider.get_raw_data_file_path(report_date)
            activities = self.activity_reader.read(file_path)

            no_activities_for_date = len(activities) == 0

            if no_activities_for_date:
                continue

            no_work_activities_for_date = not some(activities, lambda activity: activity.is_work_time)

            if no_work_activities_for_date:
                continue

            active_days_count += 1
            all_activities.extend(activities)

        holder = ActivityStatHolder(all_activities)
        matcher = self.activity_reader.matcher

        holder.initialize_stats(matcher.detailed_app_infos)
        holder.initialize_stats(matcher.distracting_app_infos)

        return holder, active_days_count

    def __get_layout(self) -> html.Div:
        """Get layout with calendar (DatePickerRange) & loading spinner & report output"""
        today = date.today()

        return html.Div(
            [
                html.H3(self.localizator.get('report_server.page_title')),
                dcc.DatePickerRange(
                    id=ElementId.DATE_PICKER.value,
                    display_format='YYYY-MM-DD',
                    first_day_of_week=1,
                    min_date_allowed=self.files_provider.get_date_of_first_raw_data_file(),
                    max_date_allowed=today + timedelta(days=1),
                    initial_visible_month=today,
                    # NOTE: allow to select the single date for start & end
                    minimum_nights=0,
                    start_date=today,
                    end_date=today,
                    updatemode='bothdates',
                ),
                dcc.Loading(
                    [html.Div(id=ElementId.REPORT_OUTPUT.value)],
                    type='graph',
                    color=choice(self.colors)
                ),
            ],
            style=dict(display='flex', flexDirection='column', alignItems='center'),
        )

    def __get_report(self, activity_stat_holder: ActivityStatHolder, active_days_count: int) -> pd.DataFrame:
        """Transform data from ActivityStatHolder to DataFrame for report"""
        report_data = {title: stat.work_time for title, stat in activity_stat_holder.items()}

        report = pd.DataFrame().from_dict(report_data, orient='index').reset_index()

        title_col = self.ColumnName.TITLE.value
        work_time_col = self.ColumnName.WORK_TIME.value
        mean_work_time_col = self.ColumnName.MEAN_WORK_TIME.value
        mean_work_time_str_col = self.ColumnName.MEAN_WORK_TIME_STR.value

        report.columns = [
            title_col,
            work_time_col,
        ]
        report.loc[report[title_col].eq(SpecialApplicationInfoTitle.BREAK_TIME.value), title_col] = \
            self.localizator.get('report_server.special_title.break_time')
        report.loc[report[title_col].eq(SpecialApplicationInfoTitle.OTHERS.value), title_col] = \
            self.localizator.get('report_server.special_title.others')

        report[work_time_col] = pd.to_timedelta(report[work_time_col])

        report = report.loc[report[work_time_col].gt(timedelta(0))]
        report[mean_work_time_col] = report[work_time_col] / active_days_count

        # from timedelta as a string extract only time
        time_re = re.compile(r'.*([\d]{2}):([\d]{2}):([\d]{2}).*')
        report[mean_work_time_str_col] = report[mean_work_time_col].astype(str).str.replace(time_re, r'\1:\2:\3')

        return report

    def __get_report_html(self, activity_stat_holder: ActivityStatHolder,
                          report: pd.DataFrame, active_days_count: int) -> html.Div:
        if report.empty:
            return html.Div(self.localizator.get('report_server.no_data'))

        figure = px.pie(report,
                        values=self.ColumnName.MEAN_WORK_TIME.value,
                        names=self.ColumnName.TITLE.value,
                        custom_data=[self.ColumnName.MEAN_WORK_TIME_STR.value],
                        color_discrete_sequence=self.colors)
        figure.update(layout_showlegend=False)

        figure.update_traces(textposition='inside',
                             hovertemplate='%{label} | %{customdata}',
                             textinfo='percent+label')

        figure.update_layout(
            hoverlabel=dict(
                bgcolor='white',
                font_size=14,
                font_family='Rockwell',
            )
        )

        distracting_app_infos = self.activity_reader.matcher.distracting_app_infos
        distracting_app_titles = [app_info.title for app_info in distracting_app_infos]

        break_time = activity_stat_holder[SpecialApplicationInfoTitle.BREAK_TIME.value].work_time
        mean_break_time = break_time / active_days_count
        mean_total_work_time = activity_stat_holder.total_work_time / active_days_count
        distracting_time = activity_stat_holder.get_group_work_time(distracting_app_titles)
        mean_distracting_time = distracting_time / active_days_count
        mean_pure_work_time = mean_total_work_time - mean_break_time - mean_distracting_time

        format_time = DatetimeFormatter.format_time_without_seconds

        return html.Div([
            html.Table(
                # headers
                [html.Tr([html.Th(col) for col in [
                    self.localizator.get('report_server.expected_col'),
                    self.localizator.get('report_server.actual_col'),
                    self.localizator.get('report_server.pure_work_col'),
                    self.localizator.get('report_server.breaks_col'),
                    self.localizator.get('report_server.distracting_col'),
                ]])] +
                # body
                [html.Tr([html.Td(time) for time in [
                    format_time(timedelta(hours=self.work_time_limit)),
                    format_time(mean_total_work_time),
                    format_time(mean_pure_work_time),
                    format_time(mean_break_time),
                    format_time(mean_distracting_time)
                ]])]
            ),
            html.Div(
                [
                    self.localizator.get('report_server.average_text'),
                    html.B(f'{active_days_count} ', style=dict(color='darkcyan')),
                    self.localizator.get('report_server.days_text'),
                ],
                style=dict(textAlign='center')
            ),
            dcc.Graph(figure=figure)
        ])

    def run(self) -> None:
        @self.app.callback(  # type: ignore[misc]
            Output(ElementId.REPORT_OUTPUT.value, 'children'),
            [
                Input(ElementId.DATE_PICKER.value, 'start_date'),
                Input(ElementId.DATE_PICKER.value, 'end_date'),
            ]
        )
        def handle_date_picker_change(start_date_value: str,
                                      end_date_value: str) -> Optional[html.Div]:
            """Update report output when date range on calendar (DatePickerRange) is changed"""
            try:
                start_date = datetime.strptime(start_date_value, '%Y-%m-%d').date()
                end_date = datetime.strptime(end_date_value, '%Y-%m-%d').date()
                report_dates = DatetimeHelper.get_dates_between(start_date, end_date)

                activity_stat_holder, active_days_count = self.__get_activity_stat_holder(report_dates)
                report = self.__get_report(activity_stat_holder, active_days_count)

                return self.__get_report_html(activity_stat_holder, report, active_days_count)

            except Exception as err:
                return html.Div(self.localizator.get('report_server.error', err=str(err)),
                                style={'textAlign': 'center'})

        self.app.run_server(self.host, self.port)



import logging
from pathlib import Path
from typing import List

from .activity import Activity
from .activity_converter import ActivityConverter
from .application_info_matcher import ApplicationInfoMatcher


class ActivityReader:
    """Read line by line activities from file with raw data"""

    def __init__(self, logger: logging.Logger, matcher: ApplicationInfoMatcher) -> None:
        self.logger = logger
        self.matcher = matcher

    def read(self, raw_data_file: Path) -> List[Activity]:
        activities: List[Activity] = []

        if not raw_data_file.exists():
            self.logger.debug(f'File with raw data [{raw_data_file}] does not exist for this day')

            return activities

        with open(str(raw_data_file)) as file:
            for line in file:
                activity = ActivityConverter.from_string(line)
                self.matcher.set_if_matched(activity)

                activities.append(activity)

        return activities



from datetime import timedelta

from .activity import Activity
from .activity_helper import ActivityHelper


class ActivityStat:
    """Store and update the amount of time spent in a certain activity"""

    def __init__(self, work_time: timedelta = timedelta(),
                 off_time: timedelta = timedelta()) -> None:
        """
        Can be used for the first reading ApplicationInfo
        from detailed/distracting lists when no activity has started yet
        """
        self.work_time = work_time
        self.off_time = off_time

    @staticmethod
    def from_activity(activity: Activity) -> 'ActivityStat':
        """For creating ActivityStat when Activity has already started"""

        ActivityHelper.raise_if_not_finished(activity)

        # NOTE: for distracting activity work_time is distracting time

        if activity.is_work_time:
            work_time = ActivityHelper.get_activity_time(activity)
            off_time = timedelta()
        else:
            work_time = timedelta()
            off_time = ActivityHelper.get_activity_time(activity)

        return ActivityStat(work_time, off_time)

    def update(self, activity: Activity) -> None:
        ActivityHelper.raise_if_not_finished(activity)

        if activity.is_work_time:
            self.work_time += ActivityHelper.get_activity_time(activity)
        else:
            self.off_time += ActivityHelper.get_activity_time(activity)

    def __eq__(self, other: object) -> bool:
        """
        Overrides the default implementation
        to use the object values instead of identifiers for comparison
        """
        if not isinstance(other, ActivityStat):
            return False

        if self.work_time != other.work_time:
            return False

        return self.off_time == other.off_time



from pathlib import Path
from typing import cast, Optional, TextIO

from pyee import BaseEventEmitter

from .activity import Activity
from .activity_converter import ActivityConverter
from .activity_helper import ActivityHelper
from .activity_splitter import ActivitySplitter
from .files_provider import FilesProvider


class ActivityWriter:
    """
    Open file that contains date in its name and write activity at string format to this file.
    If the activity lasted for several days, then it will be written in several files
    """

    FILE_MODE = 'a'
    NEW_DAY_EVENT = 'new-day-event'

    def __init__(self, files_provider: FilesProvider) -> None:
        self.__files_provider = files_provider
        self.__current_file: Optional[TextIO] = None
        self.__current_file_path: Optional[Path] = None
        self.event = BaseEventEmitter()

    def __open_file(self, file: Path) -> None:
        self.__current_file_path = file
        # cast open() result to TextIO to fix mypy warnings
        self.__current_file = cast(TextIO, open(str(file), self.FILE_MODE))

    def __write_and_flush(self, activity: Activity) -> None:
        if self.__current_file is None:
            raise Exception('current_file should be opened!')

        self.__current_file.write(ActivityConverter.to_string(activity))
        self.__current_file.flush()

    def write(self, original_activity: Activity) -> None:
        ActivityHelper.raise_if_not_finished(original_activity)

        activities_with_days = ActivitySplitter.split_by_day(original_activity)

        for (day, activity) in activities_with_days:
            file = self.__files_provider.get_raw_data_file_path(day)

            if self.__current_file_path == file:
                self.__write_and_flush(activity)
                continue

            if self.__current_file is not None:  # current day != day of opening
                self.__current_file.close()
                self.event.emit(ActivityWriter.NEW_DAY_EVENT)

            self.__open_file(file)
            self.__write_and_flush(activity)
